from .query import A2SQuery
from .data import SourceInfo, GoldSourceInfo, Player
from .enums import ServerType, Environment

__all__ = (
    "A2SQuery", "SourceInfo", "GoldSourceInfo",
    "Player", "ServerType", "Environment"
)



import struct

from .data import SourceInfo, GoldSourceInfo, Player
from .enums import ServerType, Environment

__all__ = ("Parser",)


class Parser:

    def __init__(self, data: bytes):
        self.index = 0
        self.data = data

    def __enter__(self):
        self.index = 0
        return self

    def __exit__(self, exc_val, exc_type, exc_tb):
        if not exc_val:
            return True
        return False

    def read_byte(self) -> int:
        value = struct.unpack("<B", self.data[self.index: self.index + 1])[0]
        self.index = self.index + 1

        return value

    def read_string(self) -> str:
        end = self.data.index(b"\x00", self.index)

        value = self.data[self.index: end].decode("utf-8")
        self.index = end + 1

        return value

    def read_short(self) -> int:
        value = struct.unpack("<h", self.data[self.index: self.index + 2])[0]
        self.index = self.index + 2

        return value

    def read_char(self) -> str:
        return chr(self.read_byte())

    def read_bool(self) -> bool:
        return bool(self.read_byte())

    def read_long(self) -> int:
        value = struct.unpack("<l", self.data[self.index: self.index + 4])[0]
        self.index = self.index + 4

        return value

    def read_long_long(self) -> int:
        value = struct.unpack("<Q", self.data[self.index: self.index + 8])[0]
        self.index = self.index + 8

        return value

    def read_float(self) -> float:
        value = struct.unpack("<f", self.data[self.index: self.index + 4])[0]
        self.index = self.index + 4

        return value

    @classmethod
    def parse_source_info(cls, data: bytes) -> SourceInfo:
        with cls(data) as parser:
            protocol = parser.read_byte()
            name = parser.read_string()
            info_map = parser.read_string()
            folder = parser.read_string()
            game = parser.read_string()
            app_id = parser.read_short()
            players = parser.read_byte()
            max_players = parser.read_byte()
            bots = parser.read_byte()
            server_type = ServerType(parser.read_char())
            environment = Environment(parser.read_char())
            password = parser.read_bool()
            vac = parser.read_bool()

            mode = None
            witnesses = None
            duration = None
    
            if app_id == 2400:
                mode = parser.read_byte()
                witnesses = parser.read_byte()
                duration = parser.read_byte()
    
            version = parser.read_string()
            extra_data_flag = parser.read_byte()

            port = None
            steam_id = None
            spectator_port = None
            spectator_name = None
            keywords = None
            game_id = None
    
            if extra_data_flag & 0x80:
                port = parser.read_short()
    
            if extra_data_flag & 0x10:
                steam_id = parser.read_long_long()
    
            if extra_data_flag & 0x40:
                spectator_port = parser.read_short()
                spectator_name = parser.read_string()
    
            if extra_data_flag & 0x20:
                keywords = parser.read_string()
    
            if extra_data_flag & 0x01:
                game_id = parser.read_long_long()
    
            return SourceInfo(
                protocol=protocol, name=name, map=info_map,
                folder=folder, game=game, app_id=app_id,
                players=players, max_players=max_players, bots=bots,
                server_type=server_type, environment=environment, password=password,
                vac=vac, version=version, extra_data_flag=extra_data_flag,
                mode=mode, witnesses=witnesses, duration=duration,
                port=port, steam_id=steam_id, spectator_port=spectator_port,
                spectator_name=spectator_name, keywords=keywords, game_id=game_id,
            )
            
    @classmethod
    def parse_goldsource_info(cls, data: bytes) -> GoldSourceInfo:
        with cls(data) as parser:
            address = parser.read_string()
            name = parser.read_string()
            info_map = parser.read_string()
            folder = parser.read_string()
            game = parser.read_string()
            players = parser.read_byte()
            max_players = parser.read_byte()
            protocol = parser.read_byte()
            server_type = ServerType(parser.read_char())
            environment = Environment(parser.read_char())
            password = parser.read_bool()
            modded = parser.read_bool()

            mod_link = None
            mod_download_link = None
            mod_version = None
            mod_size = None
            mod_multiplayer_only = None
            mod_uses_custom_dll = None

            if modded:
                mod_link = parser.read_string()
                mod_download_link = parser.read_string()
                mod_version = parser.read_long()
                mod_size = parser.read_long()
                mod_multiplayer_only = parser.read_bool()
                mod_uses_custom_dll = parser.read_bool()

            vac = parser.read_bool()
            bots = parser.read_byte()

        return GoldSourceInfo(
            address=address, name=name, map=info_map,
            folder=folder, game=game, players=players,
            max_players=max_players, protocol=protocol, server_type=server_type,
            environment=environment, password=password, modded=modded,
            mod_link=mod_link, mod_download_link=mod_download_link, mod_version=mod_version,
            mod_size=mod_size, mod_multiplayer_only=mod_multiplayer_only, mod_uses_custom_dll=mod_uses_custom_dll,
            vac=vac, bots=bots
        )

    @classmethod
    def parse_players(cls, data: bytes) -> list[Player]:
        players = []

        with cls(data) as parser:
            player_count = parser.read_byte()

            while len(players) < player_count:
                players.append(Player(
                    index=parser.read_byte(),
                    name=parser.read_string(),
                    score=parser.read_long(),
                    duration=parser.read_float()
                ))

            if parser.index < len(data):
                for player in players:
                    player.deaths = parser.read_long()
                    player.money = parser.read_long()

        return players

    @classmethod
    def parse_rules(cls, data: bytes) -> dict[str, str]:
        with cls(data) as parser:
            rule_count = parser.read_short()
            rules = dict((parser.read_string(), parser.read_string()) for _ in range(rule_count))

        return rules



import socket
import struct
import typing

from .data import SourceInfo, GoldSourceInfo, Player
from .exceptions import InvalidResponse, SocketClosed
from .parser import Parser
from .enums import RequestType, ResponseType, ResponseFormat

__all__ = ("QueryResponse", "A2SQuery")


class QueryResponse:

    def __init__(self, response_type: ResponseType, response_format: ResponseFormat, data: bytes):
        self.type = response_type
        self.format = response_format
        self.data = data


class A2SQuery:
    """Query various information from running Source/GoldSource game servers.

    This class allows you to interface with servers that implement the A2S query protocol.
    Each instance of A2SQuery opens a socket and connects to the specified server until closed.
    A2SQuery will authenticate server challenge requests.
    """

    def __init__(self, host: str, port: int = 27015, timeout: float = 10):
        """Create a new A2SQuery instance connected to the specified server.

        Arguments:
            host: The IP address of the server. Do not include a port here.
            port: The query port of the server. This is the same as the connection port for most games.
            timeout: How long to wait for connection/requests before timing out.
        """
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._socket.connect((host, port))
        self._socket.settimeout(timeout)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val:
            raise exc_val
        self.close()

    def close(self) -> None:
        """Close the query socket. All requests after this will fail."""
        self._socket.close()
        self._socket = None

    def _request(self, request_type: RequestType, body: str = None, challenge: int = -1) -> QueryResponse:
        if self._socket is None:
            raise SocketClosed("The socket has been closed. No more requests can be made.")

        if body is None:
            body = ""

        self._socket.send(struct.pack(f"<lB{len(body)}sl", -1, request_type.value, body.encode(), challenge))

        response_data = self._socket.recv(65536)

        parser = Parser(response_data)

        response = QueryResponse(
            response_format=ResponseFormat(parser.read_long()),
            response_type=ResponseType(parser.read_byte()),
            data=response_data[parser.index:]
        )

        if response.type is ResponseType.Challenge:
            if challenge != -1:
                raise InvalidResponse("Server requested too many challenges")

            return self._request(request_type, body, challenge=struct.unpack("<l", response.data)[0])

        return response

    def info(self) -> typing.Union[SourceInfo, GoldSourceInfo]:
        """Query general information about the server.

        Returns:
            :class:`a2squery.SourceInfo` or :class:`a2squery.GoldSourceInfo` depending on server's engine/response.
        """
        response = self._request(RequestType.Info, "Source Engine Query\x00")

        if response.type is ResponseType.InfoSource:
            return Parser.parse_source_info(response.data)
        if response.type is ResponseType.InfoGoldSource:
            return Parser.parse_goldsource_info(response.data)

        raise InvalidResponse(
            f"Invalid server response type "
            f"(got {response.type}, expected {ResponseType.InfoSource} or {ResponseType.InfoGoldSource})"
        )

    def player(self) -> list[Player]:
        """Query the server's current players/bots.

        Returns:
            List of Player objects
        """
        response = self._request(RequestType.Player)

        if response.type is ResponseType.Player:
            return Parser.parse_players(response.data)

        raise InvalidResponse(
            f"Invalid server response type "
            f"(got {response.type}, expected {ResponseType.Player})"
        )

    def players(self) -> list[Player]:
        """Query the server's current players/bots.

        This is an alias of A2SQuery.player().

        Returns:
            List of Player objects
        """
        return self.player()

    def rules(self) -> dict[str, str]:
        """Query the server's rules/configuration variables in key/value pairs.

        The Console variables included are the ones marked with FCVAR_NOTIFY
        as well as any additional ones listed in the server configuration.

        Returns:
            Key/value dictionary of rules
        """
        response = self._request(RequestType.Rules)

        if response.type is ResponseType.Rules:
            return Parser.parse_rules(response.data)

        raise InvalidResponse(
            f"Invalid server response type "
            f"(got {response.type}, expected {ResponseType.Rules})"
        )



from dataclasses import dataclass
from typing import Optional

from .enums import Environment, ServerType

__all__ = ("SourceInfo", "GoldSourceInfo", "Player")


class Data:

    def __iter__(self):
        for k in self.__slots__:
            yield k, getattr(self, k)

    def __getitem__(self, item):
        if item in self.__slots__:
            return getattr(self, item)
        raise KeyError(item)

    def keys(self):
        return self.__slots__

    def values(self):
        return [getattr(self, k) for k in self.__slots__]

    def items(self):
        return [(k, getattr(self, k)) for k in self.__slots__]

    def get(self, key, default=None):
        return getattr(self, key) or default


@dataclass(slots=True)
class SourceInfo(Data):
    """Represents a Source server's information response
    
    Attributes:
        protocol: The version of the protocol used by the server.
        name: The hostname of the server. This is what you see on a server list.
        map: The current map.
        folder: The name of the folder containing the game files. (ex. Counter-Strike: Global Offensive servers will return "csgo")
        game: The full name of the game the server is running.
        app_id: The steam App ID of the server. (ex. 730 for Counter-Strike: Global Offensive servers)
        players: The number of players on the server. This may include bots.
        max_players: The max number of players allowed.
        bots: The number of bots on the server.
        server_type: The type of server. Servers can be Dedicated, NonDedicated, or SourceTV.
        environment: The operating system the server is running. For Source servers this can be Linux, Mac, or Windows.
        password: Whether of not the server is password locked.
        vac: Indicates whether the server is protected by VAC or not.
    
        version: The version of the game installed on the server.
        extra_data_flag: This field specifies what extra data is included in the packet.
    
        mode:
            The game mode the server is currently running.

            - 0 for Hunt
            - 1 for Elimination
            - 2 for Duel
            - 3 for Deathmatch
            - 4 for VIP Team
            - 5 for Team Elimination

            .. warning::

                This field is only populated on servers running
                `The Ship: Murder Party <https://store.steampowered.com/app/2400/The_Ship_Murder_Party/>`_
                where :py:attr:`a2squery.SourceInfo.app_id` == 2400.

        witnesses:
            The number of witnesses necessary to have a player arrested.

            .. warning::

                This field is only populated on servers running
                `The Ship: Murder Party <https://store.steampowered.com/app/2400/The_Ship_Murder_Party/>`_
                where :py:attr:`a2squery.SourceInfo.app_id` == 2400.

        duration:
            The time before a player is arrested while being witnessed in seconds.

            .. warning::

                This field is only populated on servers running
                `The Ship: Murder Party <https://store.steampowered.com/app/2400/The_Ship_Murder_Party/>`_
                where :py:attr:`a2squery.SourceInfo.app_id` == 2400.

        port: The server's game port.
        steam_id: The server's steam ID.
        spectator_port: The spectator port used for SourceTV.
        spectator_name: The name of the SourceTV spectator server.
        keywords: Tags used to describe the server.
        game_id: The server's 64-bit game ID.
    """

    protocol: int
    name: str
    map: str
    folder: str
    game: str
    app_id: int
    players: int
    max_players: int
    bots: int
    server_type: ServerType
    environment: Environment
    password: bool
    vac: bool

    version: str
    extra_data_flag: int

    mode: Optional[int] = None
    witnesses: Optional[int] = None
    duration: Optional[int] = None

    port: Optional[int] = None
    steam_id: Optional[int] = None
    spectator_port: Optional[int] = None
    spectator_name: Optional[str] = None
    keywords: Optional[str] = None
    game_id: Optional[int] = None


@dataclass(slots=True)
class GoldSourceInfo(Data):
    """Represents a GoldSource server's info response.

    Attributes:
        address: The IP address and port of the server.
        name: The hostname of the server. This is what you see on a server list.
        map: The current map.
        folder: The name of the folder containing the game files. (ex. Counter-Strike: Global Offensive servers will return "csgo")
        game: The full name of the game the server is running.
        players: The number of players on the server. This may include bots.
        max_players: The max number of players allowed.
        protocol: The version of the protocol used by the server.
        server_type: The type of server. Servers can be Dedicated, NonDedicated, or SourceTV.
        environment: The operating system the server is running. For GoldSource servers this can be Linux or Windows.
        password: Whether of not the server is password locked.
        modded: Indicates if the server is a Half-Life mod or not.

        mod_link:
            A URL to the mod's website.

            .. danger::

                This field is only populated when :py:attr:`a2squery.GoldSourceInfo.modded` is True.

        mod_download_link:
            A URL to download the mod.

            .. danger::

                This field is only populated when :py:attr:`a2squery.GoldSourceInfo.modded` is True.

        mod_version:
            The version of the mod running on the server.

            .. danger::

                This field is only populated when :py:attr:`a2squery.GoldSourceInfo.modded` is True.

        mod_size:
            The space used by the mod in bytes.

            .. danger::

                This field is only populated when :py:attr:`a2squery.GoldSourceInfo.modded` is True.

        mod_multiplayer_only:
            Indicates whether the mod is multiplayer only or single-player and multiplayer.

            .. danger::

                This field is only populated when :py:attr:`a2squery.GoldSourceInfo.modded` is True.

        mod_uses_custom_dll:
            Indicates whether the mod uses its own DLL or the Half-Life DLL.

            .. danger::

                This field is only populated when :py:attr:`a2squery.GoldSourceInfo.modded` is True.


        vac: Indicates whether the server is protected by VAC or not.
        bots: The number of bots on the server.
    """

    address: str
    name: str
    map: str
    folder: str
    game: str
    players: int
    max_players: int
    protocol: int
    server_type: ServerType
    environment: Environment
    password: bool
    modded: bool

    vac: bool
    bots: int

    mod_link: Optional[str] = None
    mod_download_link: Optional[str] = None
    mod_version: Optional[int] = None
    mod_size: Optional[int] = None
    mod_multiplayer_only: Optional[bool] = None
    mod_uses_custom_dll: Optional[bool] = None


@dataclass(slots=True)
class Player(Data):
    """Represents a queried player.

    Attributes:
        index: The index of the player's chunk. This is not the player's index in the list of players.
        name: The player's name.
        score: The player's score. This is unique to the server/game but is usually kills/points.
        duration: How long the player has been connected, in seconds.
        deaths:
            The player's deaths.

            .. warning::

                This field is only populated on servers running
                `The Ship: Murder Party <https://store.steampowered.com/app/2400/The_Ship_Murder_Party/>`_
                where :py:attr:`a2squery.SourceInfo.app_id` == 2400.

        money:
            The player's deaths.

            .. warning::

                This field is only populated on servers running
                `The Ship: Murder Party <https://store.steampowered.com/app/2400/The_Ship_Murder_Party/>`_
                where :py:attr:`a2squery.SourceInfo.app_id` == 2400.
    """

    index: int
    name: str
    score: int
    duration: float

    deaths: Optional[int] = None
    money: Optional[int] = None



__all__ = ("SourceQueryException", "InvalidResponse", "SocketClosed")


class SourceQueryException(Exception):
    pass


class InvalidResponse(SourceQueryException):
    pass


class SocketClosed(SourceQueryException):
    pass



from enum import Enum

__all__ = (
    "RequestType", "ResponseType", "ResponseFormat",
    "ServerType", "Environment"
)


class RequestType(Enum):

    Info = 0x54
    Player = 0x55
    Rules = 0x56


class ResponseType(Enum):

    Challenge = 0x41
    InfoSource = 0x49
    InfoGoldSource = 0x6D
    Player = 0x44
    Rules = 0x45


class ResponseFormat(Enum):

    Simple = -1
    Batch = -2


class ServerType(Enum):

    Dedicated = "d"
    NonDedicated = "l"
    SourceTV = "p"
    Unknown = "unknown"

    @classmethod
    def _missing_(cls, value: str):
        if value.lower() != value:
            return cls(value.lower())
        return cls.Unknown


class Environment(Enum):

    Linux = "l"
    Windows = "w"
    Mac = "m"
    Unknown = "unknown"

    @classmethod
    def _missing_(cls, value: str):
        if value == "o":
            return cls.Mac
        if value.lower() != value:
            return cls(value.lower())
        return cls.Unknown
